package Lento::Bootstrap;

use strict;

use Lento::Serverdb;
use Lento::Vldb;
use Lento::Acceptor;
use Lento::Downcall;

use Lento::Debuggable;
use Getopt::Long;
use POE::Session;
use Lento::ReqDispatcher;

@Lento::Bootstrap::ISA = qw(Lento::Debuggable);

=pod 

=head1 Lento::Bootstrap

=head1 Description

The Bootstrap package is responsible for parsing lento''s command line
arguments and starting the services.

=over 4

=item  B<Set up a server database> $::serverdb which servers do we know about

=item B<Set up a volume database> $::vldb on which server can we find a volume

=item B<Set up a root volume> If a server is given as a commandline
argument, add it to the server database and add a root volume for this
server to the $::vldb.

=item  B<Set up the FSDB> $::fsdb, the database of file system objects.

=item B<Initialize kernel communication driver> If we are a client or
server, which needs presto initialize a handle for this device.  The
$::psdev object of type Psdev is structured as a POE::Driver object.
The device is opened.

=item B<Request dispatcher> A session is instantiated.

=item B<CallBackTable> Always instantiate an acceptor on port 2222.
Set up a ServerSemantics object $::CallbackTable if we are running a
server.

=back

=cut


my ($nopresto, $debugging, $root, $device);

sub args {
    # startup options
    my $debugging;
    GetOptions("d=s" => \$debugging, "root=s" => \$::root,
	       "runserver!" => \$::runserver,
	       "device=s" => \$device,
	       "nopresto!" => \$nopresto)  || die "Getoptions";

    $::global_debug = 1 if $debugging;
    # root of file system
    $::root = "/izo" unless $::root;
    $device = "/dev/loop0" unless $device;
    LOG ">>>>> ARGS: mount point: $::root, device: $device\n";
}

sub start {
    my $class = shift;
    my $self = bless {}, $class;

    new POE::Session([], 'Lento::Bootstrap' , [ '_start' , '_stop' ]);
    return undef;
}

sub _start {
    # deal with the command line
    &args();

    # Serverdb
    $::server_port = 2222;
    $::serverdb = Lento::Serverdb->new();
    #initialize volume database
    $::vldb = Lento::Vldb->new();

    # open the kernel request channel
    unless ($nopresto) {
	$::psdev = Lento::Psdev->new();
	$::psdev->open();
    }

    $::dispatcher = Lento::ReqDispatcher->new($::psdev);

    # I am a server:
    Lento::Acceptor->start($::server_port);
    if (defined($::runserver)) {
	$::CallbackTable = Lento::ServerSemantics->new();
    }

    $::psdev->poll_journal() unless $nopresto;
    system("mount $::root") unless $nopresto;

}

sub _stop {
    my $object = $_[OBJECT];
    print "$object has stopped, bye world...\n";
    exit;
}
1;
