
# Client Modification Log

package Lento::CML;
use strict;
use Lento::Debuggable;
use POE;
use POSIX qw(:errno_h);

@Lento::CML::ISA = qw(Lento::Debuggable);
$Lento::CML::level = 0;

sub new ($$) {
    my $class = shift;
    my $vol = shift;
    my $self;
    $self->{records} = [ ];
    $self->{vol} = $vol;
    $self->{seq_offset} = 0;
    bless $self, $class;

    return $self;
}

sub append ($$) {
    my $self = shift;
    my $ref = shift;
    my @records = @{$ref};

    push @{$self->{records}}, \@records;
}

#
# Method to dispatch the records to the server that controls the appropriate
# volume.
#
sub dispatch ($) {
    my $self = shift;
    my $session = shift;  # in case the session needs to be told when done

    return if (!scalar(@{$self->{records}}));

    if ($self->{offset}) {
	# Do nothing--we are already shipping, and we don't want to really
	# screw this up.

	DEBUG "Already shipping--postponing further dispatch\n";
	$self->{pending_dispatch} = 1;
	return;
    }

    $self->{offset} = scalar(@{$self->{records}});

    new POE::Session
      (
       [$self, $session], # one argument
       $::dispatcher, ['connectme'],
       '_start' => sub {
	   my $cml = $_[ARG0];
	   $_[HEAP]->{tell_when_done} = $_[ARG1];
	   $_[SESSION]->{name} = "$cml->{vol}->{name} CML";
	   $_[SESSION]->option(trace => 1, default => 1) if ($CML::level > 0);
	   $_[HEAP]->{cml} = $cml;
	   $_[KERNEL]->alias_set($_[SESSION]);
	   $_[KERNEL]->yield('connectme', $_[HEAP]->{cml});
       },
       'got_connection' => sub {
	   my $cml = $_[HEAP]->{cml} || die;

	   # Know how many records to delete once we're assured that they were
	   # successfully reintegrated.
	   DEBUG "Length: %d\n", scalar(@{$cml->{records}});

	   $cml->{vol}->{server}->{connection}->sendpacket('REQ',
			    ['Reintegrate', $cml->{vol}->name(),
			     $cml->{seq_offset}, \@{$cml->{records}}]);

	   # If pending_dispatch is defined, the CML has records pending that
	   # have not been sent to the server.  Since we're sending all records
	   # now, clear this flag.
	   delete $self->{pending_dispatch};
       },
       REP => sub {
	   # we only get a reply if there's an error--otherwise, we start
	   # getting data packets indicating successful reintegration of a few
	   # records.  Handle the error.

	   my $packet = $_[ARG0];
	   my $cml = $_[HEAP]->{cml};
	   my $rc = shift @{$packet->{parms}};

	   # XXX handle the error!
       },
       DAT => sub {
	   # the server sends us a data packet every time it completes the
	   # reintegration of some records.  Inside the packet are the
	   # sequence numbers of the records completed.
	   my $packet = $_[ARG0];
	   my $cml = $_[HEAP]->{cml};
	   my $arrayref = Storable::thaw($packet->{body});

	   while (@{$arrayref}) {
	       my $rec_no = shift @{$arrayref};

	       # the sequence offset is stored just to make sure that the
	       # server is completing them in order.  If for any reason it
	       # tells us that it finished a record out of order, we have a
	       # problem.
	       if ($rec_no != $cml->{seq_offset}) {
		   die "BAD BAD BAD--server sent invalid sequence number - " .
		     "expected $cml->{seq_offset} got $rec_no\n";
	       }

	       # remove the first record from the list, update offset
	       $cml->{seq_offset}++;
	       splice(@{$cml->{records}}, 0, 1);
	   }
       },
       EOD => sub {
	   my $packet = $_[ARG0];

	   $packet->{connection}->unuse();
	   # cleanup.
	   $_[KERNEL]->yield('done', $packet);
       },
       'done' => sub {
	   my $packet = $_[ARG0];
	   my $cml = $_[HEAP]->{cml};

	   if ( defined $_[HEAP]->{tell_when_done} ) {
	       $poe_kernel->post($_[HEAP]->{tell_when_done}, 'reint_done');
	   }


	   Lento::ACT->remove_ACT($packet->{ctoken});
	   $_[KERNEL]->yield('disconnectme', $_[HEAP]->{cml});
	   $_[KERNEL]->alias_remove($_[SESSION]);
	   delete $cml->{offset};
	   $cml->dispatch() if (defined $cml->{pending_dispatch});
       },
      );
}

1;
