package Lento::Connection;
use strict;
use Socket;
use Lento::Filter::Packet;
use POE qw(Wheel::SocketFactory Wheel::ReadWrite Driver::SysRW Filter::Line);
use Lento::Debuggable;
use Carp qw(confess cluck);
use Lento::List;
use POSIX qw(:errno_h);

@Lento::Connection::ISA = qw(Lento::Debuggable);
$Lento::Connection::level = 1;

=pod

=head1 NAME

Connections - InterMezzo Connection class

=head1 Description

Connections manage an accepted and connected TCP sockets.  When in a
connected state, each object has POE session with a Packet Wheel to
receive and send data.

This dispatching function is performed by the got_data event handler,
invoked by the Wheel and Filter.  B<Incoming> packets need a
destination session. This is determined with the routine
packet_to_session, which follows the following rules:

=over 4

=item * the ReqDispatcher [type == REQ, session == none]

=item * an already active ReqSession expecting data.  This session is
resolved (with the help of the Asynchronous Completion Token (ACT)
package) through the B<tokens> in the packet.  For REP, DAT, EOD, EOR
packets the B<ctoken> is used, and for MSG packets the B<stoken> is
used.

=back


If the PacketWheel tells us that the connection was closed, this class will:

=over 4

=item * destroy the connection session, and wheel set up by the attach
method of the Connection, and close the socket.

=item * notify the Connector with a 'got_error' event.

=back

Connections objects are attached only instantiated by Connectors and
the Acceptor.

=head1 Events propagated to other sessions

=over 4

=item I<got_wheel>: posted to the connector when the socket has been
instantiated, and the Connection session is ready for use.

=item I<got_error>: propagate operation and errnum to all users and
Connector.

=item I<REQ,REP,DAT,EOD,EOR,MSG> (MSG not yet implemented).  Propagate
these events to other sessions when packets come in.

=back 

=head1 Methods and Events


=cut

sub new {
    my $class = shift;
    my $self = bless { } , $class;
}


=pod

=over 4

=item B<attach> (Method)

Set set up a session owning the SysRW wheel for an accepted or
connected socket. 

B<Example:>  $conn->attach($socket, $connector, $peerid);

=back

=cut

sub attach { my $self = shift;

    $self->{socket} = shift;
    $self->{connector} = shift;
    $self->{peer} = shift;

    $self->{max_silence} = 5;
    $self->{monitor_delay} = 3;

    $self->{users} = new Lento::List;

    # XXX should we assign $self->{session} here or in _start ??? XXXX
    new POE::Session
      (
       [$self],
       $self, ['_start', 'got_error', 'got_packet', 'do_monitor'],
       'Lento::Connection', ['REQ', 'REP', 'DAT', 'EOD', 'EOR'],
       _stop => sub {
	   LOG "<<    $_[SESSION]->{name} going away.\n";
       },
       # all output flushed
       flushed => sub {
       }
      );
}

=pod 

=over 4

=item B<_start> (Event) 

Event is delivered by the attach method, and attached a SysRW Wheel to the accepted or connected socket.

=back

=cut

sub _start {
    my $self = $_[OBJECT];
    my $packet;
    
    $self->{session} = $_[SESSION];
    
    $_[SESSION]->{name} = "Connection ($self->{peer})";
    LOG ">>>>> Connection $self->{peer} established\n";
    
    $_[SESSION]->option(trace => 1, default => 1)
      if ($Connection::level > 0);
    
    # start the read/write wheel
    $_[HEAP]->{wheel} = new POE::Wheel::ReadWrite
      (Handle => $self->{socket},          # on this socket handle
       Driver => new POE::Driver::SysRW(), # use sysread and syswrite
       Filter => new Lento::Filter::Packet(),     # parse streams as packets
       InputState   => 'got_packet', # generate this event upon receipt
       # of input
       ErrorState   => 'got_error', # and this event if an error occurs
       FlushedState => 'flushed'    # this event when all output is sent
      );
    
    $self->{last_recv} = time();
    
    if ($self->{connector}) {
	# if we post, there is a race condition
	$_[KERNEL]->call($self->{connector}, 'got_wheel');
    }
    $_[KERNEL]->delay('do_monitor', $self->{monitor_delay});
}

=pod

=over 4

=item  B<do_monitor> (Event)

Check if the socket has not received a packet for max_silence
or more.  If so, raise and error.

=back

=cut

sub do_monitor {
    if (time() - $_[OBJECT]->{last_recv} > $_[OBJECT]->{max_silence}) {
	$_[KERNEL]->yield('got_error', 'monitor', ETIMEDOUT, 'timeout');
	LOG(">>>> $_[SESSION]->{name} dead, disconnecting.\n");
    } else {
	$_[KERNEL]->delay('do_monitor', $_[OBJECT]->{monitor_delay});
    }
}

=over 4

=item B<got_error> (event)

Arguments: operation, errnum, errstr.  Invoked by the Wheel when an
error or disconnect occurs.  The action taken is to destroy the Wheel, 
close the socket, and to remove the Connection session from the object.

If the connection was instantiated by a Connector, the Connector is
notified with a 'got_error' event.  All users of the connection are
notified.

=back

=cut
sub got_error {
    my ($operation, $errnum, $errstr) = @_[ARG0, ARG1, ARG2];
    my $self = $_[OBJECT];

    if ($errnum) { # non-zero errnum is an error
	LOG(">     $_[SESSION]->{name} detected $operation error " .
	      "$errnum: $errstr\n");
    } else { # zero errnum is a plain disconnect
	LOG ">     $_[SESSION]->{name} detected a remote disconnect.\n";
    }
    
    # either way, disconnect
    delete $_[HEAP]->{wheel};
    close $self->{socket};
    delete $self->{session};
    # must be a call to avoid race condition
    if ($_[OBJECT]->{connector}) {
	$_[KERNEL]->call($_[OBJECT]->{connector}, 'got_error',
			 $operation, $errnum);
    }

    $self->{users}->iterate(sub { DEBUG "Posting to $_[0]->{name}\n"; $poe_kernel->post(@_); }, 
			    
			    'got_error', $operation, $errnum, $errstr);

    $self->{users} = new Lento::List;
}

=pod

=over 4


=item B<got_packet> (event) 

Invoked by the Wheel/Filter when a packet arrives.

=back

=cut
sub  got_packet {
    my $packet = $_[ARG0];
    
    DEBUG("Connection=$_[SESSION]->{name}: dispatching packet: for "
	  . "token $packet->{ctoken}\n");
    
    # ignore input on lingering socket
    # return if ($_[HEAP]->{'protocol state'} eq 'quitting');
    
    $_[OBJECT]->{last_recv} = time();
    
    $packet->{connection} = $_[OBJECT];
    my $session = packet_to_session($packet);
    DEBUG("Got " . $packet->event_type() . 
	  " packet for $_[SESSION]->{name}\n");
    $_[KERNEL]->post($session, $packet->event_type(), $packet);
}

=pod

=over 4

=item B<packet_to_session> (method)

Used by the got_packet method to determine the destination of the packet.

=back

=cut
sub packet_to_session {
    my $packet = shift;
    my $session;

    if ($packet->{type} == $Packet::REQ) {
	$session = 'ReqDispatcher';
    } elsif ($packet->{type} == $Packet::MSG) {
	$session = Lento::ACT->token_to_session($packet->{stoken});
    } else {
	$session = Lento::ACT->token_to_session($packet->{ctoken});
    }
    confess unless $session;

    return $session;
}

#
#  SENDING stuff
#


=pod

=over 4

=item B<sendpacket> (method)

This function is a wrapper for the individual packet sending events
delivered to the session attached to the connection.  This wrapper is
present to handle failed connections gracefully; here the connection
object will persist, but the session may disappear, and the wrapper is
the entity that knows about this.

Any session calling this is added to the list of users.  These users
are notified if an error occurs.

B<Examples:>

$conn->sendpacket('REQ', ["method", "arg1", ...]);

$conn->sendpacket('REP', $packet, ["result", "arg1", ...]);

$conn->sendpacket('DAT', $packet, \$buffer);

$conn->sendpacket('EOD', $packet);

$conn->sendpacket('EOR', $packet);


=back

=cut
sub sendpacket {
    my $self = shift;

    if ( $self->{session} ) {
        $self->{users}->add($poe_kernel->session())
                   unless $self->{users}->exists($poe_kernel->session());
        $poe_kernel->post($self->{session}, @_);
    } else { 
        $poe_kernel->yield('got_error', ENOTCONN, 'Not Connected');
    }
}

=pod

=over 4

=item B<unuse> (method)

Called from within a session when the connection is no longer used by
the session.  Error notifications will not be delivered after this is
called.

=back

=cut

sub unuse {
    my $self = shift;

    if ( $self->{users}->exists($poe_kernel->session()) ) {
         $self->{users}->del($poe_kernel->session());
    } else {
         cluck "Trying to unuse connection which is not in use\n";
    }
}

=pod

=head1 Sending Packet events

One sends packets by posting events to a Connection session.  This is
only done from sendpacket and not directly, to avoid sending messages
to evaporated sessions.  The various packet types are detailed below,
with the appropriate event name in brackets.

=over 4

=item * Request [REQ]

Request requires at least one argument, the method to be executed on the remote
system.  Additional optional arguments are passed to that method.

B<Example:>

$_[KERNEL]->post($conn->{session}, 'REQ', ["method", "arg1", ...]);

=cut
sub REQ {
    my $args = $_[ARG0];
    my $packet = new Lento::Packet;

    $packet->{type} = $Packet::REQ;
    if (@$args == 0) {
	confess "No method in req packet\n";
    }

    DEBUG "* $_[SENDER]->{name} using $_[SESSION]->{name} to send REQ\n";
    my $act = Lento::ACT->new($_[SENDER]);
    $packet->{ctoken} = $act;
    # freeze method with args
    $packet->{body} = Storable::nfreeze($args);
    $packet->{stoken} = undef;
    $packet->{size} = length($packet->{body});
    die "Sendreq but no wheel!!!" unless $_[HEAP]->{wheel}; 
    $_[HEAP]->{wheel}->put($packet);
}

=pod

=item * Reply [REP]

Reply requires two arguments: the packet which is being replied to, and the
result code.  Additional optional arguments are passed to the reply handler.

B<Example:>

$_[KERNEL]->post($conn->{session}, 'REP', $packet, ["result", "arg1", ...]);

=cut

# pass this method: packet, result code, other return vals
sub REP {
    my $oldpkt = $_[ARG0];
    my $repargs = $_[ARG1];

    my $packet = $oldpkt->clone();
    my $result = $repargs->[0];
    if (! defined $result) {
	confess "no return code in rep packet\n";
    }

    $packet->{type} = $Packet::REP;

    $packet->{body} = Storable::nfreeze($repargs);
    $packet->{size} = length($packet->{body});
    LOG "* $_[SESSION]->{name} sending REP  %s.\n", $packet->printme();
    die "Sendreq but no wheel!!!" unless $_[HEAP]->{wheel}; 
    $_[HEAP]->{wheel}->put($packet);
}

=pod

=item * Data [DAT]

Data requires one and only one argument, the packet which is being replied to.
packet->{body} B<must> be set before one tries to send a data packet.

B<Example:>

$kernel->post($conn->{session}, 'DAT', $packet, \$buffer);

=cut

# arguments: packet.  packet must already have a valid body!
sub DAT {
    my $oldpkt = $_[ARG0];
    my $refbuf = $_[ARG1];

    my $packet = $oldpkt->clone();

    $packet->{type} = $Packet::DAT;
    $packet->{body} = $$refbuf;
    $packet->{size} = length($packet->{body});

    LOG "* $_[SESSION]->{name} sending DAT, %s.\n", $packet->printme();
    die "Trying to send packet but no wheel!!!" unless $_[HEAP]->{wheel};
    $_[HEAP]->{wheel}->put($packet);
}

=pod

=item * End of Data [EOD]

End of Data requires one and only one argument, the packet which is being
replied to.  The packet''s body, size, etc are automatically cleared.

B<Example:>

$kernel->post($conn->{session}, 'EOD', $packet);

=cut

# arguments: packet
sub EOD {
    my $oldpkt = $_[ARG0];

    my $packet = $oldpkt->clone();

    $packet->{type} = $Packet::EOD;
    $packet->{size} = 0;
    $packet->{body} = "";
    LOG "* $_[SESSION]->{name} sending EOD, %s.\n", $packet->printme();
    die "Sendreq but no wheel!!!" unless $_[HEAP]->{wheel}; 
    $_[HEAP]->{wheel}->put($packet);
}

=pod

=item * End of Request [EOR]

End of Request requires one and only one argument, the packet which is being
replied to.  The packet''s body, size, etc are automatically cleared.

B<Example:>

$kernel->post($conn->{session}, 'EOR', $packet);

=back

=cut

# arguments: packet
sub EOR {
    my $oldpkt = $_[ARG0];

    my $packet = $oldpkt->clone();

    $packet->{type} = $Packet::EOR;
    $packet->{size} = 0;
    $packet->{body} = "";
    LOG "* $_[SESSION]->{name} sending EOR.\n";
    $packet->{wheel}->put($packet);
}

1;


package TestClientConnection;
use POE;
use POE::Session;


sub _start {
  my ($object, $heap, $socket) = @_[OBJECT, HEAP, ARG0];
  my $packet;
                                        # start the read/write wheel
  $heap->{'wheel'} = new POE::Wheel::ReadWrite
    ( Handle     => $socket,                # on this socket handle
      Driver     => new POE::Driver::SysRW, # using sysread and syswrite
      Filter     => new Lento::Filter::Packet,    # and parsing streams as lines
      InputState => 'got_response',         # generate this event upon input
      ErrorState => 'got_error'             # and this event in an error occurs
    );
           

  $packet = Lento::Packet->new();
  $packet->init(1, 2, 3, "Moose\n");
  $heap->{wheel}->put($packet);

  $packet = Lento::Packet->new();
  $packet->init(1, 2, 3, "Hello\n");
  $heap->{wheel}->put($packet);

  $packet = Lento::Packet->new();
  $packet->init(1, 2, 3, "");
  $heap->{wheel}->put($packet);

}

1;

__END__
