package Lento::Connector;
use strict;
use Lento::Debuggable;
use Lento::List;
use Lento::Connection;
use Socket;
use POE;
use Lento::ACT;
use Carp qw(cluck);

@Lento::Connector::ISA = qw(Lento::Debuggable);
$Lento::Connector::level = 1;

=pod

=head1 Connectors

=head1 Description

A connector object is a daemon session which manages the Connection
session to a peer/port pair, from a client perspective.  It is used on
the systems which establish connections, not on those which accept
connections.

It knows of all the sessions that requested such a connection in a
List of clients.

The connection session is not explicitly held by the connector, to
avoid race conditions in its destruction upon errors.  The Connection
object is held.

The responsibilities of this class and session are:

=over 4

=item * to initiate a connection to a peer on behalf of a client session.

=item * to monitor such a connection with periodic Ping requests.

=item * to de-register client sessions when they are done with a connection.

=item * to notify connecting client sessions of error conditions which
led the socket to be closed.

=back

=head1 Events propagated to other sessions

=over 4

=item I<got_connection>: posted to the caller(s) of B<connectme> when the
Connection session is ready.

=item I<got_error>: propagate operation and errnum to all clients.

=back 

=head1 Events handled

=over 4

=item I<got_connection>: the connect has succeeded. 

=item I<got_wheel>: the wheel now exists. Now propagate the
got_connection event to client sessions.

=item I<got_error>: an error arose during connection setup. Setup and
propagate to clients.

=back

=head1 Mehods and Events

=over 4

=item B<new>: create a new connector. Arguments: peer, peerport.

=back

=cut

sub new ($$$) {
    my $class = shift;
    my $self = bless { }, $class;

    $self->{connection} = shift;
    $self->{peer} = shift;
    $self->{peerport} = shift;
    $self->{clients} = new Lento::List;

    DEBUG "Connector to $self->{peer} is being created.\n";

    # wrap this object in a POE session
    new POE::Session
      (
       [$self], # one argument to _start
       $self, ['get_connection', 'got_connection', 'got_wheel', 'got_error'],
       _start => sub {
	   my $object = $_[ARG0];
	   $object->{session} = $_[SESSION];
	   $_[SESSION]->{name} = "Connector (" . $object->{peer} . ")";
	   $_[SESSION]->option(trace => 1, default => 1)
	     if ($Connector::level > 0);
	   $_[KERNEL]->alias_set($object->{peer} . "_connector");
       }
      );
    
    $self->make_pinger();
    return $self;
}

=pod

=over 4

=item B<make_pinger> (Method)

Set up a periodic pinging session to poll liveness of the Peer.

=back

=cut
sub make_pinger {
    my $self = shift;
    # start a Pinger to this server
    my $ping_delay = 5;
    new POE::Session
      (
       [$self, $ping_delay], # _start args
       $self, ['connectme'],
       _start => sub {
	   $_[SESSION]->{name} = "Pinger for $_[ARG0]->{peer}";
	   $_[HEAP]->{connector} = $_[ARG0];
	   $_[HEAP]->{delay} = $_[ARG1];

	   $_[SESSION]->option(trace => 1, default => 1)
	     if ($Connector::level > 0);
	   $_[KERNEL]->alias_set($_[HEAP]->{alias}, 
				 "Pinger ($_[ARG0]->{peer})");
	   $_[KERNEL]->delay('do_connect', $_[HEAP]->{delay});
       },
       got_connection => sub {
	   my $connection = $_[HEAP]->{connector}->{connection};

	   $connection->sendpacket('REQ', ['Ping']);
       },
       REP => sub {
	   my $packet = $_[ARG0];

	   $_[KERNEL]->delay('do_connect', $_[HEAP]->{delay});
	   $_[HEAP]->{connector}->{connection}->unuse();
	   Lento::ACT->remove_ACT($packet->{ctoken});
       },
       # This is an event handler to accomodate for the delay
       do_connect => sub {
	   $_[HEAP]->{connector}->connectme();
       },
       got_error => sub {
	   $_[KERNEL]->delay('do_connect', $_[HEAP]->{delay});
	   $_[HEAP]->{connector}->{connection}->unuse();
	   $_[KERNEL]->alias_remove("Pinger ($_[HEAP]->{connector}->{peer})");
       },
       _stop => sub {
	   LOG "$_[SESSION]->{name} being stopped\n";
       }
      );
}

=pod

=over 4

=item B<get_connection> (Event)

Perform a connect to a peer, using a connecting SocketFacotry
wheel. We will be notified of success at the got_connection event.

=back

=cut

sub get_connection { 

    die "Connected already!" if ($_[OBJECT]->{connection}->{session});
    die "Connecting already!" if ($_[HEAP]->{wheel});

    # the wheel doing a connect to the peer
    $_[HEAP]->{wheel} = POE::Wheel::SocketFactory->new
      (SocketDomain    => AF_INET,    # Internet address family
       SocketType      => SOCK_STREAM, # create stream sockets
       SocketProtocol  => 'tcp',          # using the tcp protocol
       RemoteAddress   => $_[OBJECT]->{peer}, 
       RemotePort      => $_[OBJECT]->{peerport},
#        SO_RCVTIMEO     => 10,
       Reuse           => 'yes',            # reusing address and port
       SuccessState    => 'got_connection', # send when connected
       FailureState    => 'got_error',  # send this upon failure
      );
}

=pod

=over 4

=item B<connectme> (Method) 

Yield with a I<got_connection> message to the caller if an existing
Connection session exists.  Else set up a Connection session to the
peer/port.  When a new Connection has been created, all client
sessions will be notified with a I<got_connection> event in the
I<got_wheel> event handler.

=back

=cut

sub connectme {
    my $self = shift;

    if ($self->{connection}->{session}) {
	$poe_kernel->yield('got_connection');
    } else {
	DEBUG "Adding " . $poe_kernel->session()->{name} . 
	  "($poe_kernel->session()) to clients\n";
	$self->{clients}->add($poe_kernel->session());
	if ( $self->{state} ne 'connecting') {
	    $self->{state} = 'connecting';
	    $poe_kernel->post($self->{session}, 'get_connection');
	}
    }
}

=pod

=over 4

=item B<got_connection>  (Event)

The socket is now connected. Proceed by attaching the Connection to a
SysRW wheel using this socket.

=back

=cut
sub got_connection {
    my $socket = $_[ARG0];
    my $connector = $_[OBJECT];

    delete $_[HEAP]->{wheel};

    $_[OBJECT]->{connection}->attach($socket, $_[SESSION],
				     "to " . $connector->{peer});
}

=pod

=over 4

=item B<got_wheel>  (Event)

The socket with the SysRW Wheel is now ready for use.  Notify all
clients that they can start using the connection.

=back

=cut
sub got_wheel {
    $_[OBJECT]->{clients}->iterate
      (
	sub {
	    DEBUG "Notifying $_[0]->{name}: got_connection.\n";
	    $poe_kernel->post(@_);
	}, 'got_connection'
      );
    # no one is using us anymore for connecting
    $_[OBJECT]->{state} = "";
    $_[OBJECT]->{clients} = new Lento::List;
}

=pod

=over 4

=item B<got_error> (Event)

This event can be delivered by the connecting wheel or by the
Connection during connecting or socket establishment operations. It
notifies all clients of the error.

=back

=cut

sub got_error {
    my ($operation, $errnum) = @_[ARG0, ARG1];
    my $self = $_[OBJECT];
    my $client;

    $self->{clients}->iterate
      ( sub { $poe_kernel->post(@_) },
	'got_error', $self->{peer}, $operation, $errnum
      );

    $_[OBJECT]->{clients} = new Lento::List;

}

1;
