# 
# Copyright (1999) Stelias Computing Inc, braam@stelias.com
# Copyright (1999) Carnegie Mellon University, braam@cs.cmu.edu
# 
package Lento::Packet;
use Lento::Debuggable;

@Lento::Packet::ISA = qw(Lento::Debuggable);
$Lento::Packet::level = -1;
#use strict;

my $HDRSIZE = 20;  # size, ctoken, stoken, iuid, body_type

$Packet::REQ = 0x1;  # always the 1st msg from the clnt -> srv
$Packet::REP = 0x2;  # always srv -> clnt
$Packet::EOR = 0x3;  # last packet srv -> clnt, destruct session, implies EOD
$Packet::DAT = 0x4;  # srv -> clnt
$Packet::EOD = 0x5;  # srv -> clnt
$Packet::MSG = 0x6;  # cln -> srv, always sent srv -> clnt packet arrived

$Packet::str =  [undef, 'REQ', 'REP', 'EOR', 'DAT', 'EOD', 'MSG'];
my $packettype = { 'REQ' => $Packet::REQ, 'REP' => $Packet::REP, 
		   'EOR' => $Packet::EOR, 'DAT' => $Packet::DAT, 
		   'EOD' => $Packet::EOD, 'MSG' => $Packet::MSG };

sub new {
    my $class = shift;
    my $self = bless {}, $class;
    $self;
}

sub clone {
    my $oldpkt = shift;

    my $newpkt = new Lento::Packet;
    $newpkt->{ctoken} = $oldpkt->{ctoken};
    $newpkt->{stoken} = $oldpkt->{stoken};
    $newpkt->{iuid} = $oldpkt->{iuid};

    return $newpkt;
}

sub event_type {
    my $self = shift;

    return $Packet::str->[$self->{type}];
}

sub getheader {
    my $self = shift;
    my $buf = shift;
    
    ($self->{size}, $self->{ctoken}, $self->{stoken}, 
     $self->{iuid}, $self->{type} ) 
             =  unpack "NNNNN", $buf;
    DEBUG "%s\n", $self->printme();
}

sub printme {
    my $self = shift;

    return("type: $self->{type}, size: $self->{size}, ctoken: $self->{ctoken}"
	   . ", stoken: $self->{stoken}, iuid: $self->{iuid}");
}

# class method
sub init {
    my $self = shift;
    my ($type, $ctoken, $stoken, $iuid, $body) = @_;

    my $size = length($body);
    $self->{type} = $type;
    $self->{size} = $size;
    $self->{body} = $body;
    $self->{ctoken} = $ctoken;
    $self->{stoken} = $stoken;
    $self->{iuid} = $iuid;
    $self;
}

sub packme {
    my $self = shift;
#    my $size = $self->{size};
    my $size = length($self->{body});
    pack ("NNNNNa$size", $self->{size}, $self->{ctoken}, 
	  $self->{stoken}, $self->{iuid}, $self->{type}, 
	  $self->{body});
}

sub len {
    my $self = shift;
    return $self->{size};
}

sub adddata {
    my $self = shift;
    my $buf = shift;
    my $offset = shift;
    $self->{body} = substr($buf, $offset, $self->{size});
}

sub getbody {
    my $self = shift;
    my $buf = shift;

    die "Read error!" if $self->{size} < length($buf);
    $self->{body} = substring($buf, $HDRSIZE, $self->{size});
}


1;

package Lento::Filter::Packet;
use Storable qw(nfreeze thaw);
use Lento::ACT;
use strict;
use Lento::Debuggable;

@Lento::Filter::Packet::ISA = qw(Lento::Debuggable);
$Lento::Filter::Packet::level = -1;
#------------------------------------------------------------------------------

sub new {
  my $type = shift;
  my $self = bless { }, $type;
  $self->{basetype} = $type;

  $self;
}

sub put {
    my ($self, $packets) = @_;
    my @raw;
    foreach my $packet (@$packets) {
	# print $packet->printme();
	push(@raw, $packet->packme());
    }

    \@raw;
}

# XXXX Can be done more efficiently by processing the individual elements
# rather than concatenating the buffer.
#
sub get {
    my ($self, $stream) = @_;
    my @result;
    my $left;
    my $offset;


    $self->{buffer} .= join('', @$stream);
    # adjust $left whenever we truncate the buffer
    $left = length($self->{buffer});
    my $bufsize = $left;
    
    while ( $left >= 0 ) {
	DEBUG "GET: left $left, bufsize $bufsize\n";
	if ( !$self->{packet} ) {
	    $self->{havehdr} = 0;
	    $self->{packet} = new Lento::Packet();
	}

	# have no header, want one
	if ( $left >= $HDRSIZE && !$self->{havehdr} ) {
	    DEBUG "Adding header: left $left, bufsize $bufsize\n";
	    $self->{packet}->getheader($self->{buffer});
	    $self->{havehdr} = 1;
	    $offset  += $HDRSIZE;
	    $self->{buffer} = substr($self->{buffer}, $offset);
	    $left -= $offset;
	    $offset = 0;
	}

	# nothing useful can be grabbed
	if ( !$self->{havehdr} ) {
	    DEBUG "No header: last, left $left, bufsize $bufsize, offset $offset\n";
	    last;
	}
	
	# here we know we have a header

	# no body is following
	if ($self->{packet}->{size} == 0) {
	    # remove the head from the buffer
	    DEBUG "No size: left $left, bufsize $bufsize\n";
	    $self->{buffer} = substr($self->{buffer}, $offset);
	    $left -= $offset;
	    $offset = 0;

	    push(@result, $self->{packet});
	    $self->{havehdr} = 0;
	    delete $self->{packet};
	    next;
	} 

	# we have a full body; add it to the result
	if  ( $left >= $self->{packet}->{size} ) {
	    DEBUG "--->left $left, bufsize $bufsize, offset $offset\n";
	    $self->{packet}->adddata($self->{buffer}, $offset);
	    $offset += $self->{packet}->{size};
	    DEBUG "--->left $left, bufsize $bufsize, offset $offset\n";

	    # remove the head from the buffer
	    $self->{buffer} = substr($self->{buffer}, $offset);
	    $left -= $offset;
	    $offset = 0;
	    DEBUG "--->left $left, bufsize $bufsize, offset $offset\n";
	    DEBUG $self->{packet}->printme();
	    DEBUG("LEFT: $left\n");

	    # only REQ and REP need to have the body thawed
	    if ( $self->{packet}->{type} == $Packet::REQ ||
	         $self->{packet}->{type} == $Packet::REP ) {
		my $refin = Storable::thaw $self->{packet}->{body};
		$self->{packet}->{parms} = $refin;
	    }

	    push(@result, $self->{packet});
	    $self->{havehdr} = 0;
	    delete $self->{packet};
	    next;
	}
	last;
    }
    # we will only trigger the receiver if a header is all we get
    \@result;
}

1;
