# unpacking kernel arguments
#use strict;
use POSIX qw(:errno_h);
package Lento::Upcall;
use Lento::Psdev;
use Lento::Debuggable;
use POE;

@Lento::Filter::Upcall::ISA = qw(Lento::Debuggable);
$Lento::Filter::Upcall::level = 1;

# opcodes
my %op;
my $PERMIT=  1;   $op{$PERMIT}="Permit";
my $JOURNAL= 2;   $op{$JOURNAL}="Journal";
my $OPENDIR= 3;   $op{$OPENDIR}="OpenDir";
my $OPEN=    4;   $op{$OPEN}="OpenFile";
my $SIGNAL=  5;   $op{$SIGNAL}="Signal";

# Class Method
sub new ($$) {
    shift; # class
    my $buf = shift;
    my $self;

    ($self->{opc}, $self->{uniq}, $self->{pid}, $self->{uid}) =
      unpack "iiss", $buf;

    die "Upcall::new, opcode $self->{opc} not known" unless
      defined $op{$self->{opc}};

    &{"Lento::Upcall::$op{$self->{opc}}::init"}($self, $buf, 12);

    DEBUG $self->printme;
    return $self;
}

# Method
sub op {
    my $self = shift;
    $op{$self->{opc}};
}


# Method
sub printme {
    my $self = shift;

    return "$self->{class}: $op{$self->{opc}} [uniq $self->{uniq} pid " .
      "$self->{pid} uid $self->{uid}] path: $self->{path} nm1: $self->{nm1} " .
      "nm2: $self->{nm2}\n";
}

# method
# replies to the kernel won't block: do them synchronously
sub complete {
    my $self = shift;
    my $result = shift;

    my $answer = pack "iii", $self->{opc}, $self->{uniq}, $result;

    # mount does opendir on / and is async, so is Journal: don't send a reply
    return if $self->{async}; 

    DEBUG "Upcall result: $result\n";

    my $retval = $::psdev->write($answer, length($answer));
    if ($retval <= 0 && $retval != -ESRCH) {
	die "Psdev->write";
    }
}

#
# Subclasses for each upcall TYPE
#
# Upcall::Permit
package Lento::Upcall::Permit;
use strict;
use Lento::Debuggable;

@Lento::Upcall::Permit::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::Permit::level = 1;

#
# XXXX Phil, why is this test for $self->{path} eq '/' 
# that should never happen
#


sub init($$$) {
    my ($self, $buf, $offset) = @_;
    $self->{pathlen} = unpack "x$offset i", $buf;
    $self->{path} = substr($buf, $offset + 4, $self->{pathlen});

    if ($self->{path} eq "/") {
	$self->{path} = $::root;
    }

    DEBUG "Permit path: $self->{path}\n";
    bless $self;
}

# Upcall::Opendir
package Lento::Upcall::OpenDir;
use strict;
use Lento::Debuggable;

@Lento::Upcall::OpenDir::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::OpenDir::level = 1;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    ($self->{async}, $self->{pathlen}) = unpack "x$offset ii", $buf;
    $self->{path} = substr($buf, $offset + 8, $self->{pathlen});

    if ($self->{path} eq "/") {
	$self->{path} = $::root;
    }

    DEBUG "OpenDir path: $self->{path}\n";
    bless $self;
}

# Upcall::OpenFile
package Lento::Upcall::OpenFile;
use strict;
use Lento::Debuggable;
use POE;
use POSIX qw(:fcntl_h);

@Lento::Upcall::OpenFile::ISA = qw(Lento::Debuggable Lento::Upcall);
$Lento::Upcall::OpenFile::level = 1;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    ($self->{pathlen}) = unpack "x$offset i", $buf;
    $self->{path} = substr($buf, $offset + 4, $self->{pathlen});

    if ($self->{path} eq "/") {
	$self->{path} = $::root;
    }

    DEBUG "Openfile path: $self->{path}\n";
    bless $self;
}

sub got_connection {
    my $vol = $_[HEAP]->{completion}->{vol};
    my $connection = $vol->{server}->{connection};

    $_[HEAP]->{path} = $vol->fs_path($_[HEAP]->{completion}->{path});

    $connection->sendpacket('REQ', 
			    ['SendFile', $vol->name(), $_[HEAP]->{path}]);
}

sub REP {
    my $packet = $_[ARG0];
    my $rc = shift @{$packet->{parms}};

    if ($rc != 0) {
	$poe_kernel->yield('done', $rc, $packet->{ctoken});
    } else {
	$_[HEAP]->{path} =
	  $_[HEAP]->{completion}->{vol}->cache_path($_[HEAP]->{path});

	$_[HEAP]->{stat} = shift @{$packet->{parms}};
	DEBUG "Installing file $_[HEAP]->{path}\n";
	$_[HEAP]->{handle} = new IO::File;
	$_[HEAP]->{handle}->open($_[HEAP]->{path}, O_WRONLY) || die;
    }
}

sub done {
    my $vol = $_[HEAP]->{completion}->{vol};
    my $connection = $vol->{server}->{connection};

    $connection->unuse();
    Lento::ACT->remove_ACT($_[ARG1]);
    $_[HEAP]->{completion}->complete($_[ARG0]);
    $_[KERNEL]->alias_remove($_[SESSION]);
}

#
# Upcall::Journal
# This call does not always need completion to the kernel
#
package Lento::Upcall::Journal;
use strict;
use Lento::Debuggable;

@Lento::Upcall::Journal::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::Journal::level = 1;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    my $vol;
    $self->{async} = 1;
    $self->{buf} = substr($buf, $offset);
    bless $self;
}

# Upcall::Signal
package Lento::Upcall::Signal;
use strict;
use Lento::Debuggable;
@Lento::Upcall::Signal::ISA = qw(Lento::Upcall Lento::Debuggable);
$Lento::Upcall::Signal::level = 0;

sub init ($$$) {
    my ($self, $buf, $offset) = @_;
    bless $self;
}
1;

#
#  The filter for the Upcall Wheel
#
package Lento::Filter::Upcall;
#use Upcall;

sub new {
  my $type = shift;
  my $self = bless { }, $type;
  $self->{basetype} = $type;

  $self;
}

sub put {
    my ($self, $packets) = @_;
    my @raw;
    foreach my $packet (@$packets) {
	print $packet->printme();
	push(@raw, $packet->packme());
    }

    \@raw;
}

sub get {
    my ($self, $stream) = @_;
    my @result;

    foreach my $buf (@$stream) {
	my $upcall = Lento::Upcall->new($buf);
	push @result, $upcall;
    }
    \@result;
}

1;
