package Lento::InterMezzo::ReqHandler;
use Lento::Debuggable;
use POE;
use Lento::ReqDispatcher;
use Lento::Server;
use File::stat;
use Lento::Reintegrate;
use POSIX qw(:fcntl_h :errno_h);

@Lento::InterMezzo::ReqHandler::ISA = qw(Lento::Debuggable);
$Lento::InterMezzo::ReqHandler::level = 1;

# ARG0 is the packet, ARG1 is the name
sub _start {
    $_[SESSION]->{name} = $_[ARG1];
    $_[SESSION]->option(trace => 1, default => 1) if
      ($Lento::InterMezzo::ReqHandler::level > 0);
    $_[HEAP]->{req_packet} = $_[ARG0];

    # XXX this really belongs in ReqDispatcher::REQ!
    # NOTE: ACT::new sets an alias for this session, and ACT::remove_ACT
    # removes it.  DO NOT SET ANOTHER ALIAS UNLESS YOU CLEAN IT UP YOURSELF!
    $_[HEAP]->{req_packet}->{stoken} = new Lento::ACT($_[SESSION]);

    $_[KERNEL]->yield('REQ');
}


sub _stop  {
    LOG "<<    $poe_kernel->session()->{name} is done\n";
}

sub finish {
    Lento::ACT->remove_ACT($_[HEAP]->{req_packet}->{stoken});
    $poe_kernel->alias_remove($_[HEAP]->{req_packet}->{stoken});
    delete $_[HEAP]->{req_packet};
}

sub got_error {
    print "Timeout received\n";
    $poe_kernel->yield('finish');
}

sub Ping {
    my $packet = shift;

    new POE::Session
      (
       [$packet, "Ping"], 
       'Lento::InterMezzo::ReqHandler', ['_start', '_stop', 
					 'finish', 'got_error'],
       'REQ' => sub { 
	   my $pkt = $_[HEAP]->{req_packet};
	   $pkt->{connection}->sendpacket('REP', $pkt, [0]);
	   $pkt->{connection}->unuse();
	   $poe_kernel->yield('finish', [$pkt, 0]);
       }
      );
}

sub SendDir {
    my $packet = shift;

    new POE::Session
      (
       [$packet, "SendDir"], # two arguments to _start
       'Lento::InterMezzo::ReqHandler', ['_start', '_stop', 'finish'],
       'REQ' => sub {
	   # notify permit holder, await integration...
	   no strict 'refs';
	   my $volname = @{$_[HEAP]->{req_packet}->{parms}}[0];
	   my $rel_path = @{$_[HEAP]->{req_packet}->{parms}}[1];
	   my $vol;

	   if (!defined($vol = $::vldb->find_by_name($volname))) {
	       # Invalid volume name sent by client
	       $_[KERNEL]->yield('send_rep', EINVAL);
	   }

	   my $fshl_path = $vol->fshl_path($rel_path);

	   $_[SESSION]->{name} .= " ($fshl_path)";
	   DEBUG "SendDir: argument $fshl_path\n";
	   $_[KERNEL]->yield('package_dir', $fshl_path);
       },
       'package_dir' => sub {
	   my $error;

	   ($error, $_[HEAP]->{buffer}) = Lento::Cache->pack_dir($_[ARG0]);
	   $_[KERNEL]->yield('send_rep', $error);
       },
       'send_rep' => sub {
	   my $error = $_[ARG0];
	   my $pkt = $_[HEAP]->{req_packet};

	   DEBUG "Error is $error\n";
	   $pkt->{connection}->sendpacket('REP', $pkt, [$error]);
	   if ( $error ) {
	       $_[KERNEL]->yield('complete');
	   } else {
	       $_[KERNEL]->yield('send_data');
	   }
       },
       'send_data' => sub {
	   my $pkt = $_[HEAP]->{req_packet};

	   $pkt->{connection}->sendpacket('DAT', $pkt, \$_[HEAP]->{buffer});
	   $pkt->{connection}->sendpacket('EOD', $pkt);
	   $_[KERNEL]->yield('complete');
       }, 
       'complete' => sub {
	   $_[KERNEL]->yield('finish');
       }
      );
}


sub Permit {
    my $packet = shift;

    new POE::Session
      (
       [$packet, "Permit"], # two arguments to _start
       'Lento::InterMezzo::ReqHandler', ['_start', '_stop', 'finish'],
       REQ => sub {
	   my $volname = @{$_[HEAP]->{req_packet}->{parms}}[0];
	   my $vol;

	   if (!defined($vol = $::vldb->find_by_name($volname))) {
	       # Invalid volume name sent by client
	       $_[KERNEL]->yield('send_rep', EINVAL);
	   }
	   $_[HEAP]->{vol} = $vol;
	   $_[HEAP]->{volname} = $volname;
	   $_[SESSION]->{name} .= " ($volname)";
	   DEBUG "Permit: argument $volname\n";
	   $_[KERNEL]->yield('break_permit');
       },
       break_permit => sub {
	   my $vol =  $_[HEAP]->{vol};
	   my $volname = $_[HEAP]->{volname};

	   # break it
	   if ( $vol->{permit} ) {
	       my $connection = $vol->{permit}->{connection};
	       $connection->sendpacket('REQ', ['BreakPermit', $vol->name()]);
	       $_[HEAP]->{state} ='break_requested';
	   } else {
	       $_[KERNEL]->yield('REP');
	   }
       },
       REP => sub {  # get here when $vol->{permit} has reintegrated
	   $_[HEAP]->{state} ='break_done';

	   my $pkt = $_[HEAP]->{req_packet};
	   # grant the permit to calling client
	   $_[HEAP]->{vol}->{permit} = $pkt->{connection};
	   $pkt->{connection}->sendpacket('REP', $pkt, [0]);
	   $_[KERNEL]->yield('finish');
       },
       got_error => sub {
	   if ( $_[HEAP]->{state} eq 'break_requested' ) {
	       $_[KERNEL]->yield('REP');
	   }
       }
      );
}

sub BreakPermit {
    my $packet = shift;

    new POE::Session
      (
       [$packet, "BreakPermit"], # two arguments to _start
       'Lento::InterMezzo::ReqHandler', ['_start', '_stop', 'finish'],
       REQ => sub {
	   my $volname = @{$_[HEAP]->{req_packet}->{parms}}[0];
	   my $vol;

	   if (!defined($vol = $::vldb->find_by_name($volname))) {
	       # Invalid volume name sent by client
	       $_[KERNEL]->yield('send_rep', EINVAL);
	   }
	   $_[HEAP]->{vol} = $vol;
	   $_[HEAP]->{volname} = $volname;
	   $_[SESSION]->{name} .= " ($volname)";
	   DEBUG "Permit: argument $volname\n";
	   $_[KERNEL]->yield('reintegrate');
       },
       reintegrate => sub {
	   my $vol =  $_[HEAP]->{vol};
	   my $volname = $_[HEAP]->{volname};

	   # set kernel permit bit to 0
	   Downcall->revoke_permit($vol->{mtpt});

	   # pull the log
	   Downcall->release();

	   # reintegrate
	   $vol->{cml}->dispatch($_[SESSION]);

       },
       reint_done => sub {
	   my $pkt = $_[HEAP]->{req_packet};

	   $pkt->{connection}->sendpacket('REP', $pkt, [0]);
	   $_[KERNEL]->yield('finish');
       },
       got_error => sub {
	   if ( $_[HEAP]->{state} eq 'break_requested' ) {
	       $_[KERNEL]->yield('REP');
	   }
       }
      );
}

sub SendFile {
    my $packet = shift;

    new POE::Session
      (
       [$packet, "SendFile"], # two arguments to _start
       'Lento::InterMezzo::ReqHandler', ['_start', '_stop', 'finish'],
       'REQ' => sub {
	   # notify permit holder, await integration...
	   my $volname = @{$_[HEAP]->{req_packet}->{parms}}[0];
	   my $rel_path = @{$_[HEAP]->{req_packet}->{parms}}[1];
	   my $is_backfetch = @{$_[HEAP]->{req_packet}->{parms}}[2];
	   my $vol;

	   if (!defined($vol = $::vldb->find_by_name($volname))) {
	       # Invalid volume name sent by client
	       $_[KERNEL]->yield('send_rep', EINVAL);
	   }

	   DEBUG "vol name: $vol->{name} rel path: $rel_path\n";
	   my $fshl_path;

	   if ($is_backfetch) {
	       $fshl_path = $vol->cache_path($rel_path);
	       DEBUG "IS BACKFETCH ($rel_path, $fshl_path)\n";
	   } else {
	       $fshl_path = $vol->fshl_path($rel_path);
	   }

	   $_[SESSION]->{name} .= " ($fshl_path)";
	   DEBUG "SendFile: argument $fshl_path\n";
	   $_[KERNEL]->yield('prepare_file', $fshl_path);
       },
       'prepare_file' => sub {
	   my $path = $_[ARG0];
	   my $error = 0;

	   DEBUG "Path: $path\n";

	   my $st = stat($path);
	   if (!$st) {
	       $_[KERNEL]->yield('send_rep', ENOENT);
	       return;
	   }
	   if ($st->mode() & 040000) {
	       # this file is a directory
	       $_[KERNEL]->yield('send_rep', EISDIR);
	       return;
	   }

	   $_[HEAP]->{handle} = new IO::File;
	   if (!$_[HEAP]->{handle}->open($path, O_RDONLY)) {
	       # access denied, or similar
	       $_[KERNEL]->yield('send_rep', EPERM);
	       return;
	   }

	   $_[HEAP]->{offset} = 0;
	   $_[KERNEL]->yield('send_rep', 0, $st);
       },
       'send_rep' => sub {
	   my $error = $_[ARG0];
	   my $pkt = $_[HEAP]->{req_packet};
	   my $st = $_[ARG1];

	   DEBUG "Error is $error\n";
	   $pkt->{connection}->sendpacket('REP', $pkt, [$error, $st]);
	   if ($error) {
	       $_[KERNEL]->yield('complete');
	   } else {
	       $_[KERNEL]->yield('send_data', $st);
	   }
       }, 
       'send_data' => sub {
	   my $st = $_[ARG0];
	   my $pkt = $_[HEAP]->{req_packet};
	   my $buffer;

	   if ($_[HEAP]->{offset} < $st->size()) {
	       my $len = $_[HEAP]->{handle}->read($buffer, 64*1024);
	       if (!$len) {
		   die "read failed";
	       } else {
		   $_[HEAP]->{offset} += $len;
	       }

	       $pkt->{connection}->sendpacket('DAT', $pkt, \$buffer);
	       $_[KERNEL]->yield('send_data', $st);
	   } else {
	       $pkt->{connection}->sendpacket('EOD', $pkt);
	       $_[KERNEL]->yield('complete');
	   }
       },
       'complete' => sub {
	   if ($_[HEAP]->{handle}) {
	       $_[HEAP]->{handle}->close();
	       delete $_[HEAP]->{handle};
	   }
	   $_[KERNEL]->yield('finish');
       },
      );
}


sub Reintegrate {
    my $packet = shift;

    new POE::Session
      (
       [$packet, "Reintegrate"], # two arguments to _start
       'Lento::InterMezzo::UpcallHandler', ['DAT', 'EOD'],
       'Lento::InterMezzo::ReqHandler', ['_start', '_stop', 'finish'],
       'REQ' => sub {
	   my $volname = shift @{$_[HEAP]->{req_packet}->{parms}};

	   if (!defined($_[HEAP]->{vol} = $::vldb->find_by_name($volname))) {
	       # Invalid volume name sent by client
	       $_[KERNEL]->yield('send_rep', EINVAL);
	   }

	   $_[HEAP]->{seq_no} = shift @{$_[HEAP]->{req_packet}->{parms}};
	   $_[HEAP]->{records} = shift @{$_[HEAP]->{req_packet}->{parms}};

	   my $last_rec = $_[HEAP]->{vol}->{last_record};

	   if (defined($last_rec) && $last_rec >= $_[HEAP]->{seq_no}) {
	       splice(@{$_[HEAP]->{records}}, 0,
		      $last_rec - $_[HEAP]->{seq_no} + 1);
	       $_[HEAP]->{completed} = [$_[HEAP]->{seq_no} .. $last_rec];
	   } else {
	       $_[HEAP]->{completed} = [ ];
	   }

	   $_[KERNEL]->yield('do_one_record');
       },
       'do_one_record' => sub {
	   if (!@{$_[HEAP]->{records}}) {
	       # all done reintegrating!

	       if ($_[HEAP]->{completed}) {
		   # send confirmation to the server about these records
		   $_[KERNEL]->yield('send_completed_records');
	       }
	       $_[KERNEL]->yield('complete');

	       return;
	   }

	   my $r = shift(@{$_[HEAP]->{records}});

	   DEBUG "record: " . join(", ", @$r) . "\n";
	   my $func = pop(@{$r});

	   if (!&{"Lento::Reintegrate::$func"}($_[HEAP]->{vol}, @{$r})) {
	       DEBUG "Error: $!, $@\n";
	       # XXX do something about this error
	   } elsif ($func ne "CLOSE") {
	       $_[KERNEL]->yield('complete_record');
	   } else {
	       $_[HEAP]->{path} = @{$r}[2];
	       $_[KERNEL]->yield('do_backfetch');
	   }
       },
       'complete_record' => sub {
	   # note that we've completed this record
	   push(@{$_[HEAP]->{completed}}, $_[HEAP]->{seq_no});
	   $_[HEAP]->{seq_no}++;

	   if (scalar(@{$_[HEAP]->{completed}}) >= 8) {
	       # send confirmation to the server about these records
	       $_[KERNEL]->yield('send_completed_records');
	   }
	   $_[KERNEL]->yield('do_one_record');
       },
       'send_completed_records' => sub {
	   my $pkt = $_[HEAP]->{req_packet};
	   my $buf = Storable::nfreeze($_[HEAP]->{completed});

	   $pkt->{connection}->sendpacket('DAT', $pkt, \$buf);
	   $_[HEAP]->{completed} = [ ];
       },
       'do_backfetch' => sub { # initiates backfetching
	   my $vol = $_[HEAP]->{vol} || die;
	   my $pkt = $_[HEAP]->{req_packet};

	   DEBUG "reintegrating path: $_[HEAP]->{path}\n";
	   $pkt->{connection}->sendpacket('REQ',
			    ['SendFile', $vol->name(), $_[HEAP]->{path},
			     "do_backfetch"]);
       },
       'REP' => sub { # used during backfetching
	   my $packet = $_[ARG0];
	   my $rc = shift @{$packet->{parms}};

	   if ($rc != 0) {
	       # XXX there was some sort of error, handle it
	       $poe_kernel->yield('done', $rc, $packet->{ctoken});
	   } else {
	       $_[HEAP]->{path} =
		 $_[HEAP]->{vol}->fshl_path($_[HEAP]->{path});

	       $_[HEAP]->{stat} = shift @{$packet->{parms}};
	       DEBUG "Installing file $_[HEAP]->{path}\n";
	       $_[HEAP]->{handle} = new IO::File;
	       $_[HEAP]->{handle}->open($_[HEAP]->{path}, O_WRONLY | O_CREAT)
		 || die;
	   }
       },
       'done' => sub {
	   delete $_[HEAP]->{path};
	   Lento::ACT->remove_ACT($_[ARG1]);
	   $_[KERNEL]->yield('complete_record');
       },
       'complete' => sub {
	   my $pkt = $_[HEAP]->{req_packet};

	   $pkt->{connection}->sendpacket('EOD', $pkt);
	   $_[KERNEL]->yield('finish');
       },
       'got_error' => sub {
	   $_[HEAP]->{vol}->{last_record} = $_[HEAP]->{seq_no};

	   # XXX if we are backfetching, clean up the data
	   $_[KERNEL]->yield('finish');
       }
      );
}

1;

# "except for perl.  you could drop a duck on your keyboard and get running
# perl code." -- Zach Brown
