# =============================================================================
#
# There are _NO_ class methods in this module, only static methods!
#
# =============================================================================

package Lento::Journal;
use strict;
use File::stat;
use Lento::Debuggable;

@Lento::Journal::ISA = qw(Lento::Debuggable);
$Lento::Journal::level = -1;

# These are almost certainly not XP-safe!
my $ATTR_MODE = 1;
my $ATTR_UID = 2;
my $ATTR_GID = 4;
my $ATTR_SIZE = 8;
my $ATTR_ATIME = 16;
my $ATTR_MTIME = 32;
my $ATTR_CTIME = 64;

my %opcode_unpack = (1 => \&unpack_create,
		     2 => \&unpack_mkdir,
		     3 => \&unpack_unlink,
		     4 => \&unpack_rmdir,
      		     5 => \&unpack_close,
		     6 => \&unpack_symlink,
		     7 => \&unpack_rename,
		     8 => \&unpack_setattr,
		     9 => \&unpack_link,
		    10 => \&unpack_open);

sub round_len {
    my $len = shift;
    return ($len + 3) & ~0x3;
}

sub unpack_create {
    my ($buf, $pos) = @_;
    # don't need file's device
    my ($mode, $old_mtime, $old_size, $pathlen) = unpack("x$pos iiii", $buf);
    my $path = substr($buf, $pos + 16, $pathlen);

    DEBUG "CREATE: mode: $mode, mtime: $old_mtime, size: $old_size, path: \"$path\"\n";

    my $vol = $::vldb->find_by_path($path) || die;

    return ($vol, $mode, $old_mtime, $old_size, $vol->fs_path($path),
	    "CREATE");
}

sub unpack_setattr {
    my ($buf, $pos) = @_;
    my ($old_mtime, $old_size, $uid, $gid, $size, $atime, $mtime, $ctime,
	$mode, $flags, $valid, $pathlen) = unpack("x${pos}iiiiiiiiiiii", $buf);
    my $path = substr($buf, $pos + 48, $pathlen);

    DEBUG("SETATTR: mtime: $old_mtime, size: $old_size, UID: $uid, GID: $gid, "
	  . "size: $size, atime: $atime, mtime: $mtime, ctime: $ctime, mode: "
	  . "$mode, flags: $flags, valid: $valid, path: \"$path\"\n");

    my $vol = $::vldb->find_by_path($path) || die;

    return ($vol, $old_mtime, $old_size, $uid, $gid, $size, $atime, $mtime,
	    $ctime, $mode, $flags, $valid, $vol->fs_path($path), "SETATTR");
}

sub unpack_open {
    my ($buf, $pos) = @_;
    my ($old_mtime, $old_size, $pathlen) = unpack("x${pos}iii", $buf);
    my $path = substr($buf, $pos + 12, $pathlen);

    DEBUG "OPEN: mtime: $old_mtime, size: $old_size, path: \"$path\"\n";

    my $vol = $::vldb->find_by_path($path) || die;

    return ( );

#    return ($vol, $old_mtime, $old_size, $path, "OPEN");
}

sub unpack_mkdir {
    my ($buf, $pos) = @_;
    my ($old_mtime, $old_size, $mode, $pathlen) = unpack("x${pos}iiii", $buf);
    my $path = substr($buf, $pos + 16, $pathlen);

    DEBUG "MKDIR: mtime: $old_mtime, size: $old_size, mode: $mode, path: \"$path\"\n";

    my $vol = $::vldb->find_by_path($path) || die;

    return ($vol, $old_mtime, $old_size, $mode, $vol->fs_path($path), "MKDIR");
}

sub unpack_unlink {
    my ($buf, $pos) = @_;
    my ($old_mtime, $old_size, $pathlen, $namelen) =
      unpack("x${pos}iiii", $buf);
    my $path = substr($buf, $pos + 16, $pathlen);
    my $name = substr($buf, $pos + 16 + round_len($pathlen), $namelen);

    DEBUG "UNLINK: mtime: $old_mtime, size: $old_size, path: \"$path\", name: \"$name\"\n";

    my $vol = $::vldb->find_by_path($path) || die;

    return ($vol, $old_mtime, $old_size, $vol->fs_path($path), $name,
	    "UNLINK");
}

sub unpack_rmdir {
    my ($buf, $pos) = @_;
    my ($old_mtime, $old_size, $pathlen, $namelen) =
      unpack("x${pos}iiii", $buf);
    my $path = substr($buf, $pos + 16, $pathlen);
    my $name = substr($buf, $pos + 16 + round_len($pathlen), $namelen);

    DEBUG "RMDIR: mtime: $old_mtime, size: $old_size, path: \"$path\", name: \"$name\"\n";

    my $vol = $::vldb->find_by_path($path) || die;

    return ($vol, $old_mtime, $old_size, $vol->fs_path($path), $name, "RMDIR");
}

sub unpack_symlink ($$) {
    my ($buf, $pos) = @_;
    my ($old_mtime, $old_size, $pathlen, $targetlen) =
      unpack("x${pos}iiii", $buf);
    my $path = substr($buf, $pos + 16, $pathlen);
    my $target = substr($buf, $pos + 16 + round_len($pathlen), $targetlen);

    DEBUG "SYMLINK: mtime: $old_mtime, size: $old_size, path: \"$path\", target: \"$target\"\n";

    my $vol = $::vldb->find_by_path($path) || die;

    return ($vol, $old_mtime, $old_size, $vol->fs_path($path), $target,
	    "SYMLINK");
}

sub unpack_rename {
    my ($buf, $pos) = @_;
    my ($old_src_mtime, $old_src_size, $old_tar_mtime, $old_tar_size,
	$srcpathlen, $tarpathlen) = unpack("x${pos}iiiiii", $buf);

    my $srcpath = substr($buf, $pos + 24, $srcpathlen);
    my $tarpath = substr($buf, $pos + 24 + round_len($srcpathlen),
			 $tarpathlen);

    DEBUG("RENAME: old src mtime: $old_src_mtime, old src size: $old_src_size"
	  . ", old tar mtime: $old_tar_mtime, old tar size: $old_tar_size, "
	  . "srcpath: \"$srcpath\", tarpath: \"$tarpath\"\n");

    # XXX if srcpath and tarpath aren't on the same volume, must use trickery
    my $vol = $::vldb->find_by_path($srcpath) || die;

    return ($vol, $old_src_mtime, $old_src_size, $old_tar_mtime, $old_tar_size,
	    $vol->fs_path($srcpath), $vol->fs_path($tarpath), "RENAME");
}

sub unpack_link {
    my ($buf, $pos) = @_;
    my ($old_tar_dir_mtime, $old_tar_dir_size, $srcpathlen, $pathlen) =
      unpack("x${pos}iiii", $buf);
    my $srcpath = substr($buf, $pos + 16, $srcpathlen);
    my $path = substr($buf, $pos + 16 + round_len($srcpathlen), $pathlen);

    DEBUG("LINK: old target dir mtime: $old_tar_dir_mtime, "
	  . "old target dir size: $old_tar_dir_size, srcpath: \"$srcpath\", "
	  . "path: \"$path\"\n");

    my $vol = $::vldb->find_by_path($path) || die;

    return ($vol, $old_tar_dir_mtime, $old_tar_dir_size,
	    $vol->fs_path($srcpath), $vol->fs_path($path), "LINK");
}

sub unpack_close {
    my ($buf, $pos) = @_;
    my ($old_mtime, $old_size, $pathlen) = unpack("x${pos}iii", $buf);
    my $path = substr($buf, $pos + 12, $pathlen);

    DEBUG "CLOSE: mtime: $old_mtime, size: $old_size, path: \"$path\"\n";

    my $vol = $::vldb->find_by_path($path) || die;

    return ($vol, $old_mtime, $old_size, $vol->fs_path($path), "CLOSE");
}

sub Unpack {
    my $buffer = shift;

    my ($len, $pid, $uid, $fsuid, $opcode);
    my @records;

    my $length  = length($buffer);

    DEBUG "buffer length: $length\n";

    for (my $pos = 0; $pos < length($buffer); $pos += $len) {
	($len, $pid, $uid, $fsuid, $opcode) = unpack("x$pos iiiii", $buffer);

	if (!$len) {
	    DEBUG "Zero length record!  Danger?\n";
	    last;
	} elsif ($len > length($buffer) - $pos) {
	    DEBUG "Record is larger than the remaining buffer, danger!\n";
	}

	my @rec = ($len, $pid, $uid, $fsuid, $opcode);
	DEBUG "Journal opcode: $opcode pos: $pos len $len\n";
	my $func = $opcode_unpack{$opcode};

	if (!defined $func) {
	    DEBUG "Unknown opcode: $opcode\n";
	    next;
	}

	my @retval = &{$func}($buffer, $pos + 5*4);

	if (scalar(@retval)) {
	    my $vol = shift(@retval);
	    $vol->{cml}->append(\@retval);
	}
    }
}

1;
