
# Performs reintegration of CML entries on the server

package Lento::Reintegrate;
use strict;
use Lento::Debuggable;
use File::stat;
use POSIX qw(:unistd_h);

@Lento::Reintegrate::ISA = qw(Lento::Debuggable);
$Lento::Reintegrate::level = 0;

# see the appropriate warning about these in Journal.pm
my $ATTR_MODE = 1;
my $ATTR_UID = 2;
my $ATTR_GID = 4;
my $ATTR_SIZE = 8;
my $ATTR_ATIME = 16;
my $ATTR_MTIME = 32;
my $ATTR_CTIME = 64;

sub CREATE {
    my $vol = shift || die;
    my ($mode, $old_mtime, $old_size, $path) = @_;
    my $filename = $vol->fshl_path($path);

    DEBUG "Executing: open(>$filename); close(); chmod $mode\n";

    open(FILE, ">$filename") || die;
    close(FILE) || die;
    chmod $mode, $filename || die;

    return 1;
}

sub SETATTR {
    my $vol= shift || die;
    my ($old_mtime, $old_size, $uid, $gid, $size, $atime, $mtime, $ctime,
	$mode, $flags, $valid, $path) = @_;
    my $filename = $vol->fshl_path($path);
    my $st = stat($filename) || goto error;

    my $msg =  "[SETATTR: Setting:";
    if ($valid & $ATTR_UID && $valid & $ATTR_GID) {
	$msg .= " uid ($uid) gid ($gid)";
	chown $uid, $gid, $filename || goto error;
    } elsif ($valid & $ATTR_UID) {
	$msg .= " uid ($uid)";
	chown $uid, $st->gid(), $filename || goto error;
    } elsif ($valid & $ATTR_GID) {
	$msg .= " gid ($gid)";
	chown $st->uid(), $gid, $filename || goto error;
    }

    if ($valid & $ATTR_ATIME && $valid & $ATTR_MTIME) {
	$msg .= " atime ($atime) mtime ($mtime)";
	utime $atime, $mtime, $filename || goto error;
    } elsif ($valid & $ATTR_ATIME) {
	$msg .= " atime ($atime)";
	utime $atime, $st->mtime(), $filename || goto error;
    } elsif ($valid & $ATTR_MTIME) {
	$msg .= " mtime ($mtime)";
	utime $st->atime(), $mtime, $filename || goto error;
    }

    if ($valid & $ATTR_MODE) {
	$msg .= " mode ($mode)";
	chmod $mode, $filename || goto error;
    }

    if ($valid & $ATTR_SIZE && !($st->mode() & 040000)) {
	$msg .= " size ($size)";
	if ($size > 0) {
	    open FILE, "> $filename" || goto error;
	    seek(FILE, $size -1, &SEEK_SET) || goto error;
	    print FILE "0";
	    close FILE;
	} else {
	    truncate ($filename, 0) || goto error;
	}
    }

    DEBUG "$msg]\n";

    return 1;

  error:
    DEBUG "$msg Error: $!]\n";

    die;
}

sub MKDIR {
    my $vol = shift || die;
    my ($old_mtime, $old_size, $mode, $path) = @_;
    my $dirname = $vol->fshl_path($path);

    DEBUG "Executing: mkdir($dirname, $mode)\n";
    mkdir($dirname, $mode) || die;

    return 1;
}

sub UNLINK {
    my $vol = shift || die;
    my ($old_mtime, $old_size, $path, $name) = @_;
    my $filename = $vol->fshl_path($path) . "/" . $name;

    DEBUG "Executing: unlink($filename)\n";
    unlink($filename) || die;

    return 1;
}

sub RMDIR {
    my $vol = shift || die;
    my ($old_mtime, $old_size, $path, $name) = @_;
    my $filename = $vol->fshl_path($path) . "/" . $name;

    DEBUG "Executing: rmdir($filename)\n";
    rmdir($filename) || die;

    return 1;
}

sub SYMLINK {
    my $vol = shift || die;
    my ($old_mtime, $old_size, $path, $name) = @_;
    my $linkname = $vol->fshl_path($path);

    DEBUG("Executing: symlink($name, $linkname)\n");
    symlink($name, $linkname) || die;

    return 1;
}

sub RENAME {
    my $vol = shift || die;
    my ($old_src_mtime, $old_src_size, $old_tar_mtime, $old_tar_size,
	$srcpath, $tarpath) = @_;
    my $oldfile = $vol->fshl_path($srcpath);
    my $newfile = $vol->fshl_path($tarpath);

    DEBUG("Executing: rename($oldfile, $newfile)\n");
    rename($oldfile, $newfile) || die;

    return 1;
}

sub LINK {
    return undef;
}

sub CLOSE {
    my $vol = shift || die;
    my ($old_mtime, $old_size, $path) = @_;
    my $filename = $vol->fshl_path($path);

    DEBUG "Backfetching $filename\n";

    return 1;
}

1;
