package Lento::ReqDispatcher;
use Lento::Debuggable;
use POE qw(Wheel::ReadWrite Driver::SysRW);
use Socket;
use Lento::Filter::Upcall;
use Lento::Filter::Packet;
use Lento::InterMezzo::ReqHandler;
use Lento::InterMezzo::UpcallHandler;
use Lento::Volume;
use Lento::Server;
use Lento::ACT;

@Lento::ReqDispatcher::ISA = qw(Lento::Debuggable);
$Lento::ReqDispatcher::level = 0;

=pod

=head1 NAME  - ReqDispatcher

=head1 Description

Dispatches a network or upcall request, usually to a new session.
Owns the upcall wheel and handles kernel requests entirely.


=head1 Events handled

=over 4

=item  B<got_upcall>: uses an upcall wheel to get data from the kernel

=item  B<REQ>: network requests from Connections

=back

=item  the requests it gets need to be sent to

=over 8

=item - an already active ReqSession expecting data

=item - a new session

=back

=item  If the PacketWheel tells us that the connection was closed, this class will:

=over 8

=item  - destroy all sessions set up by this ServerConnection

=item  - ask the ServerSemantics to remove the client

=item - self destruct

=back

=back

=cut

sub new {
    my ($class, $psdev, $test) = @_;
    my $self = bless { } , $class;

    $self->{psdev} = $psdev;

    new POE::Session( $self,
		      ['_start', '_signal',
		       'REQ', 'got_upcall', 
		       'got_error', 'flushed' ], 
		    );
    $self;
}

sub _start {
    my ($object, $heap, $session) = @_[OBJECT, HEAP, SESSION];

    $_[SESSION]->{name} = "ReqDispatcher";

    # start the read/write wheel
    if ($object->{psdev}) {
	$heap->{'wheel'} = new POE::Wheel::ReadWrite
	  ( Handle  => $object->{psdev}->handle(), # on this socket handle
	    Driver  => new POE::Driver::SysRW(), # use sysread and syswrite
	    Filter  => new Lento::Filter::Upcall(), # parse streams as packets
	    InputState   => 'got_upcall', # generate this upon receipt of input
	    ErrorState   => 'got_error',  # and this event if an error occurs
	    FlushedState => 'flushed'  # and this event when all output is sent
	  );
    }

    $session->option(trace => 1, default => 1) if ($ReqDispatcher::level > 0);
    $_[KERNEL]->alias_set("ReqDispatcher");
}

=pod

=over 4

=item B<_signal> (Event)

Let us know what signal we got!

=back

=cut

sub _signal {
    my $signal = $_[ARG0];
    DEBUG "++> $_[SESSION]->{name} got signal $signal\n";
}


=pod

=over 4

=item B<REQ> (Event)

This event is sent by a Connection when a REQ packet has come in. It
instantiates a new sessions to handle the request.

=back

=cut

sub REQ {
    my ($object, $heap, $packet) = @_[OBJECT, HEAP, ARG0];

    my $method = shift @{$packet->{parms}};
    my $session = &{"Lento::InterMezzo::ReqHandler::$method"}($packet);
    # XXX the stoken that is in InterMezzo::ReqHandler::_start really
    # belongs here!
}

# Process an upcall

sub got_upcall {
    my ($object, $heap, $upcall) = @_[OBJECT, HEAP, ARG0];

    if ($object->{test}) {
	DEBUG $upcall->printme();
	my $method = $upcall->op();
	if ($method eq "Journal") {
 	    use Lento::Journal;
	    Lento::Journal::Unpack($upcall->{buf});
	}
    } else {
	# Method
	my $method = $upcall->op();
	$upcall->{session} = &{"Lento::InterMezzo::UpcallHandler::$method"}($upcall);
    }
}

sub connectme {
    my ($session, $completion) = @_[SESSION, ARG0];

    # XXXX  this is supposed to make a decision which session
    # will finish the upcall request.  For now just plough on:

    # make sure we have a connection to the server(s)
    die unless ($completion && $completion->{vol});
    die unless $completion->{vol}->{server};

    $completion->{vol}->{server}->{connector}->connectme();
}


#------------------------------------------------------------------------------
# Handle I/O errors.

sub got_error {
  my ($object, $heap, $operation, $errnum, $errstr) = 
    @_[OBJECT, HEAP, ARG0, ARG1, ARG2];

  print "$object detected $operation error $errnum: $errstr\n";

  # The SocketFactory wheel is the only thing keeping this session
  # alive.  Deleting it causes its destructor to be called.  The
  # session has nothing further to do after the socketfactory is
  # destroyed, so the kernel stops it.

  delete $heap->{'wheel'};
}

#------------------------------------------------------------------------------
# When all output is flushed, check to see if the client requested a
# quit.  If they did, honor it.

sub flushed {
  my $heap = $_[HEAP];

  # Deletes the read/write wheel, causing the session to stop out of
  # boredom.

  if ($heap->{'protocol state'} eq 'quitting') {
    delete $heap->{'wheel'};
  }
}

1;
